@@ -13,7 +13,8 @@
 # See the Licence for the specific language governing permissions and
 # limitations under the Licence, but note it is strictly for non-commercial use.
 
-from typing import TYPE_CHECKING, Dict, Iterable, List, Optional, Tuple, Union
+from typing import TYPE_CHECKING, Dict, Iterable, List, Optional, Tuple
+from logging import warning
 
 import numpy as np
 import qiskit.providers.aer.extensions.snapshot_expectation_value
@@ -28,13 +29,18 @@ from pytket.circuit import BasisOrder, Circuit, OpType, Node, Bit, Qubit
 from pytket.backends import Backend, CircuitNotRunError, ResultHandle, CircuitStatus
 from pytket.backends.resulthandle import _ResultIdTuple
 from pytket.backends.backendresult import BackendResult
-from pytket.device import Device, GateError, GateErrorContainer
+
+from pytket.device import Device, QubitErrorContainer
 from pytket.passes import (
     BasePass,
     RebaseIBM,
     SequencePass,
     SynthesiseIBM,
-    FullMappingPass,
+    CXMappingPass,
+    SequencePass,
+    DecomposeBoxes,
+    FullPeepholeOptimise,
+    CliffordSimp,
 )
 from pytket.pauli import Pauli, QubitPauliString
 from pytket.predicates import (
@@ -42,6 +48,7 @@ from pytket.predicates import (
     GateSetPredicate,
     NoClassicalControlPredicate,
     NoFastFeedforwardPredicate,
+    NoSymbolsPredicate,
     Predicate,
 )
 from pytket.qiskit import tk_to_qiskit
@@ -54,6 +61,8 @@ from pytket.utils.results import (
     KwargTypes,
 )
 
+from collections import defaultdict
+import itertools
 from .ibm_utils import _convert_result, _STATUS_MAP
 
 
@@ -64,16 +73,15 @@ def _default_q_index(q: Qubit) -> int:
 
 
 class _AerBaseBackend(Backend):
-    """Common base class for all Aer simulator backends
-    """
+    """Common base class for all Aer simulator backends"""
 
     _persistent_handles = False
 
     def __init__(self, backend_name: str):
         super().__init__()
         self._backend = Aer.get_backend(backend_name)
-        self._joblog: Dict[str, AerJob] = {}
         self._noise_model = None
+        self._characterisation = None
         self._device = None
         self._memory = False
 
@@ -81,6 +89,10 @@ class _AerBaseBackend(Backend):
     def _result_id_type(self) -> _ResultIdTuple:
         return (str, int)
 
+    @property
+    def characterisation(self) -> Optional[dict]:
+        return self._characterisation
+
     @property
     def device(self) -> Optional[Device]:
         return self._device
@@ -103,18 +115,20 @@ class _AerBaseBackend(Backend):
         )
         job = self._backend.run(qobj, noise_model=self._noise_model)
         jobid = job.job_id()
-        self._joblog[jobid] = job
         handle_list = [ResultHandle(jobid, i) for i in range(len(circuit_list))]
         for handle in handle_list:
-            self._cache[handle] = {}
+            self._cache[handle] = {"job": job}
         return handle_list
 
-    def _retrieve_job(self, jobid: str) -> AerJob:
-        return self._joblog[str(jobid)]
+    def cancel(self, handle: ResultHandle) -> None:
+        job = self._cache[handle]["job"]
+        cancelled = job.cancel()
+        if not cancelled:
+            warning(f"Unable to cancel job {handle[0]}")
 
     def circuit_status(self, handle: ResultHandle) -> CircuitStatus:
         self._check_handle_type(handle)
-        ibmstatus = self._retrieve_job(handle[0]).status()
+        ibmstatus = self._cache[handle]["job"].status()
         return CircuitStatus(_STATUS_MAP[ibmstatus], ibmstatus.value)
 
     def get_result(self, handle: ResultHandle, **kwargs: KwargTypes) -> BackendResult:
@@ -123,16 +137,14 @@ class _AerBaseBackend(Backend):
         except CircuitNotRunError:
             jobid, _ = handle
             try:
-                job = self._retrieve_job(jobid)
+                job = self._cache[handle]["job"]
             except KeyError:
                 raise CircuitNotRunError(handle)
 
             res = job.result()
             backresults = _convert_result(res)
-            self._cache.update(
-                (ResultHandle(jobid, circ_index), {"result": backres})
-                for circ_index, backres in enumerate(backresults)
-            )
+            for circ_index, backres in enumerate(backresults):
+                self._cache[ResultHandle(jobid, circ_index)]["result"] = backres
 
             return self._cache[handle]["result"]
 
@@ -143,7 +155,15 @@ class _AerBaseBackend(Backend):
         valid_check: bool = True,
     ) -> complex:
         if valid_check:
-            self._check_all_circuits([circuit])
+            self._check_all_circuits([circuit], nomeasure_warn=False)
+
+        circ_qbs = circuit.qubits
+        q_indices = (_default_q_index(q) for q in circ_qbs)
+        if not all(q_ind == i for q_ind, i in zip(q_indices, range(len(circ_qbs)))):
+            raise ValueError(
+                "Circuit must act on default register Qubits, contiguously from 0 onwards."
+                + f"Circuit qubits were: {circ_qbs}"
+            )
         qc = tk_to_qiskit(circuit)
         qc.snapshot_expectation_value("snap", hamiltonian, qc.qubits)
         qobj = assemble(qc)
@@ -153,9 +173,13 @@ class _AerBaseBackend(Backend):
         ]
 
     def get_pauli_expectation_value(
-        self, state_circuit: Circuit, pauli: QubitPauliString, valid_check: bool = True,
+        self,
+        state_circuit: Circuit,
+        pauli: QubitPauliString,
+        valid_check: bool = True,
     ) -> complex:
         """Calculates the expectation value of the given circuit using the built-in Aer snapshot functionality
+        Requires a simple circuit with default register qubits.
 
         :param state_circuit: Circuit that generates the desired state :math:`\\left|\\psi\\right>`.
         :type state_circuit: Circuit
@@ -179,6 +203,7 @@ class _AerBaseBackend(Backend):
         valid_check: bool = True,
     ) -> complex:
         """Calculates the expectation value of the given circuit with respect to the operator using the built-in Aer snapshot functionality
+        Requires a simple circuit with default register qubits.
 
         :param state_circuit: Circuit that generates the desired state :math:`\\left|\\psi\\right>`.
         :type state_circuit: Circuit
@@ -209,9 +234,14 @@ class _AerStateBaseBackend(_AerBaseBackend):
     def required_predicates(self) -> List[Predicate]:
         return _pure_aer_predicates
 
-    @property
-    def default_compilation_pass(self) -> BasePass:
-        return SynthesiseIBM()
+    def default_compilation_pass(self, optimisation_level: int = 1) -> BasePass:
+        assert optimisation_level in range(3)
+        if optimisation_level == 0:
+            return SequencePass([DecomposeBoxes(), RebaseIBM()])
+        elif optimisation_level == 1:
+            return SequencePass([DecomposeBoxes(), SynthesiseIBM()])
+        else:
+            return SequencePass([DecomposeBoxes(), FullPeepholeOptimise()])
 
     def process_circuits(
         self,
@@ -225,8 +255,10 @@ class _AerStateBaseBackend(_AerBaseBackend):
         )
         for handle, circ in zip(handles, circuits):
             perm = circ.implicit_qubit_permutation()
+            phase = circ.phase
+            self._cache[handle]["global_phase"] = phase
             if not all(key == val for key, val in perm.items()):
-                self._cache[handle] = {"implicit_perm_qubits": perm}
+                self._cache[handle]["implicit_perm_qubits"] = perm
         return handles
 
     def get_result(self, handle: ResultHandle, **kwargs: KwargTypes) -> BackendResult:
@@ -235,10 +267,8 @@ class _AerStateBaseBackend(_AerBaseBackend):
                 return self._cache[handle]["result"]
 
         self._check_handle_type(handle)
-
-        jobid, _ = handle
         try:
-            job = self._retrieve_job(jobid)
+            job = self._cache[handle]["job"]
         except KeyError:
             raise CircuitNotRunError(handle)
 
@@ -267,6 +297,18 @@ class _AerStateBaseBackend(_AerBaseBackend):
                     backres._unitary = permute_basis_indexing(
                         backres._unitary.T, tuple(permutation)
                     ).T
+            phase = self._cache[newhandle]["global_phase"]
+            try:
+                phase = float(phase)
+                expphase = np.exp(phase * np.pi * 1j)
+                if backres._state is not None:
+                    backres._state *= expphase
+                if backres._unitary is not None:
+                    backres._unitary *= expphase
+            except ValueError:
+                warning(
+                    "Global phase is dependent on a symbolic parameter, so cannot adjust for phase"
+                )
             self._cache[newhandle]["result"] = backres
 
         return self._cache[handle]["result"]
@@ -284,31 +326,57 @@ class AerBackend(_AerBaseBackend):
         :type noise_model: Optional[NoiseModel], optional
         """
         super().__init__("qasm_simulator")
-        self._noise_model = noise_model
-        if noise_model:
-            self._device = _process_model(noise_model)
 
+        if not noise_model or all(
+            value == [] for value in noise_model.to_dict().values()
+        ):
+            self._noise_model = None
+        else:
+            self._noise_model = noise_model
+            self._characterisation = _process_model(noise_model)
+
+            self._device = Device(
+                self._characterisation.get("NodeErrors", {}),
+                self._characterisation.get("EdgeErrors", {}),
+                self._characterisation.get("Architecture", Architecture([])),
+            )
         self._memory = True
 
     @property
     def required_predicates(self) -> List[Predicate]:
-        pred_list = [GateSetPredicate(_aer_ops.union({OpType.Measure}))]
+        pred_list = [
+            NoSymbolsPredicate(),
+            GateSetPredicate(_aer_ops.union({OpType.Measure})),
+        ]
         if self._noise_model and self._device:
             pred_list.append(ConnectivityPredicate(self._device))
         return pred_list
 
-    @property
-    def default_compilation_pass(self) -> BasePass:
-        if self._noise_model and self._device:
-            return SequencePass(
-                [
-                    RebaseIBM(),
-                    FullMappingPass(self._device, NoiseAwarePlacement(self._device)),
-                    SynthesiseIBM(),
-                ]
-            )
+    def default_compilation_pass(self, optimisation_level: int = 1) -> BasePass:
+        assert optimisation_level in range(3)
+        passlist = [DecomposeBoxes()]
+        if optimisation_level == 0:
+            passlist.append(RebaseIBM())
+        elif optimisation_level == 1:
+            passlist.append(SynthesiseIBM())
         else:
-            return SynthesiseIBM()
+            passlist.append(FullPeepholeOptimise())
+        if self._noise_model and self._device:
+            passlist.append(
+                CXMappingPass(
+                    self._device,
+                    NoiseAwarePlacement(self._device),
+                    directed_cx=True,
+                    delay_measures=False,
+                )
+            )
+            if optimisation_level == 0:
+                passlist.append(RebaseIBM())
+            elif optimisation_level == 1:
+                passlist.append(SynthesiseIBM())
+            else:
+                passlist.extend([CliffordSimp(False), SynthesiseIBM()])
+        return SequencePass(passlist)
 
     def process_circuits(
         self,
@@ -325,14 +393,65 @@ class AerBackend(_AerBaseBackend):
             raise ValueError("Parameter n_shots is required")
         return super().process_circuits(circuits, n_shots, valid_check, **kwargs)
 
+    def get_pauli_expectation_value(
+        self,
+        state_circuit: Circuit,
+        pauli: QubitPauliString,
+        valid_check: bool = True,
+    ) -> complex:
+        """Calculates the expectation value of the given circuit using the built-in Aer snapshot functionality.
+        Requires a simple circuit with default register qubits, and no noise model.
+
+        :param state_circuit: Circuit that generates the desired state :math:`\\left|\\psi\\right>`.
+        :type state_circuit: Circuit
+        :param pauli: Pauli operator
+        :type pauli: QubitPauliString
+        :param valid_check: Explicitly check that the circuit satisfies all required predicates to run on the backend. Defaults to True
+        :type valid_check: bool, optional
+        :return: :math:`\\left<\\psi | P | \\psi \\right>`
+        :rtype: complex
+        """
+        if self._noise_model:
+            raise RuntimeError(
+                "Snapshot based expectation value not supported with noise model. Use shots."
+            )
+
+        return super().get_pauli_expectation_value(state_circuit, pauli, valid_check)
+
+    def get_operator_expectation_value(
+        self,
+        state_circuit: Circuit,
+        operator: QubitPauliOperator,
+        valid_check: bool = True,
+    ) -> complex:
+        """Calculates the expectation value of the given circuit with respect to the operator using the built-in Aer snapshot functionality
+        Requires a simple circuit with default register qubits, and no noise model.
+
+        :param state_circuit: Circuit that generates the desired state :math:`\\left|\\psi\\right>`.
+        :type state_circuit: Circuit
+        :param operator: Operator :math:`H`.
+        :type operator: QubitPauliOperator
+        :param valid_check: Explicitly check that the circuit satisfies all required predicates to run on the backend. Defaults to True
+        :type valid_check: bool, optional
+        :return: :math:`\\left<\\psi | H | \\psi \\right>`
+        :rtype: complex
+        """
+        if self._noise_model:
+            raise RuntimeError(
+                "Snapshot based expectation value not supported with noise model. Use shots."
+            )
+
+        return super().get_operator_expectation_value(
+            state_circuit, operator, valid_check
+        )
+
 
 class AerStateBackend(_AerStateBaseBackend):
     _supports_state = True
     _supports_expectation = True
 
     def __init__(self):
-        """Backend for running simulations on the Qiskit Aer Statevector simulator.
-        """
+        """Backend for running simulations on the Qiskit Aer Statevector simulator."""
         super().__init__("statevector_simulator")
 
 
@@ -340,22 +459,20 @@ class AerUnitaryBackend(_AerStateBaseBackend):
     _supports_unitary = True
 
     def __init__(self):
-        """Backend for running simulations on the Qiskit Aer Unitary simulator.
-        """
+        """Backend for running simulations on the Qiskit Aer Unitary simulator."""
         super().__init__("unitary_simulator")
 
     def get_unitary(
         self,
-        circuit: Union[Circuit, ResultHandle],
+        circuit: Circuit,
         basis: BasisOrder = BasisOrder.ilo,
         valid_check: bool = True,
     ) -> np.ndarray:
         """
-        Calculate the unitary matrix for a circuit, or return unitary from
-        handle to previously processed circuit.
+        Calculate the unitary matrix for a circuit.
 
-        :param circuit: Circuit to execute or ResultHandle to fetch results for.
-        :type circuit: Union[Circuit, ResultHandle]
+        :param circuit: Circuit to execute
+        :type circuit: Circuit
         :param basis: Toggle between ILO-BE (increasing lexicographic order of
             bit ids, big-endian) and DLO-BE (decreasing lexicographic order,
             big-endian) for column ordering. Defaults to BasisOrder.ilo.
@@ -367,7 +484,7 @@ class AerUnitaryBackend(_AerStateBaseBackend):
         :rtype: np.ndarray
         """
 
-        result, _ = self._get_or_process(circuit, valid_check=valid_check)
+        result, _ = self._process(circuit, valid_check=valid_check)
         q_bits = (
             sorted(result.q_bits.keys(), reverse=(basis is not BasisOrder.ilo))
             if result.q_bits
@@ -394,8 +511,8 @@ _aer_ops = {
     OpType.Tdg,
     OpType.CCX,
     OpType.SWAP,
-    OpType.Unitary1qBox,
     OpType.Unitary2qBox,
+    OpType.Reset,
 }
 
 _pure_aer_predicates = [
@@ -405,7 +522,7 @@ _pure_aer_predicates = [
 ]
 
 
-def _process_model(noise_model: NoiseModel) -> Device:
+def _process_model(noise_model: NoiseModel) -> dict:
     # obtain approximations for gate errors from noise model by using probability of "identity" error
     _gate_str_2_optype = {
         "u1": OpType.U1,
@@ -414,9 +531,22 @@ def _process_model(noise_model: NoiseModel) -> Device:
         "cx": OpType.CX,
         "id": OpType.noop,
     }
-    errors = [e for e in noise_model.to_dict()["errors"] if e["type"] == "qerror"]
+
+    supported_single_optypes = {OpType.U1, OpType.U2, OpType.U3, OpType.noop}
+    errors = [
+        e
+        for e in noise_model.to_dict()["errors"]
+        if e["type"] == "qerror" or e["type"] == "roerror"
+    ]
     link_ers_dict = {}
-    node_ers_dict = {}
+    node_ers_dict = defaultdict(lambda: QubitErrorContainer(supported_single_optypes))
+    readout_errors_dict = {}
+    generic_single_qerrors_dict = defaultdict(lambda: list())
+    generic_2q_qerrors_dict = defaultdict(lambda: list())
+
+    node_ers_qubits = set()
+    link_ers_qubits = set()
+
     coupling_map = []
     for error in errors:
         name = error["operations"]
@@ -424,44 +554,74 @@ def _process_model(noise_model: NoiseModel) -> Device:
             raise RuntimeWarning("Error applies to multiple gates.")
         if "gate_qubits" not in error:
             raise RuntimeWarning(
-                "Please define NoiseModel without using the add_all_qubit_quantum_error() method."
+                "Please define NoiseModel without using the add_all_qubit_quantum_error() or add_all_qubit_readout_error() method."
             )
         name = name[0]
+
         qubits = error["gate_qubits"][0]
+        node_ers_qubits.add(qubits[0])
         gate_fid = error["probabilities"][-1]
         if len(qubits) == 1:
-            error_cont = GateErrorContainer(
-                {_gate_str_2_optype[name]: GateError(1 - gate_fid, 1)}
-            )
-            node_ers_dict[qubits[0]] = error_cont
+            if error["type"] == "qerror":
+                node_ers_dict[qubits[0]].add_error(
+                    (_gate_str_2_optype[name], 1 - gate_fid)
+                )
+                generic_single_qerrors_dict[qubits[0]].append(
+                    (error["instructions"], error["probabilities"])
+                )
+            elif error["type"] == "roerror":
+                node_ers_dict[qubits[0]].add_readout(error["probabilities"][0][1])
+                readout_errors_dict[qubits[0]] = error["probabilities"]
+            else:
+                raise RuntimeWarning("Error type not 'qerror' or 'roerror'.")
         elif len(qubits) == 2:
-            error_cont = GateErrorContainer(
-                {_gate_str_2_optype[name]: GateError(1 - gate_fid, 1)}
-            )
+            # note that if multiple multi-qubit errors are added to the CX gate, the resulting noise channel is composed and reflected in probabilities
+            error_cont = QubitErrorContainer({_gate_str_2_optype[name]: 1 - gate_fid})
+            link_ers_qubits.add(qubits[0])
+            link_ers_qubits.add(qubits[1])
             link_ers_dict[tuple(qubits)] = error_cont
             # to simulate a worse reverse direction square the fidelity
-            rev_error_cont = GateErrorContainer(
-                {_gate_str_2_optype[name]: GateError(1 - gate_fid ** 2, 1)}
+            rev_error_cont = QubitErrorContainer(
+                {_gate_str_2_optype[name]: 1 - gate_fid ** 2}
             )
             link_ers_dict[tuple(qubits[::-1])] = rev_error_cont
+            generic_2q_qerrors_dict[tuple(qubits)].append(
+                (error["instructions"], error["probabilities"])
+            )
             coupling_map.append(qubits)
 
-    if len(coupling_map) == 0:
-        return None
-    arc = Architecture(coupling_map)
+    free_qubits = node_ers_qubits - link_ers_qubits
+
+    for q in free_qubits:
+        for lq in link_ers_qubits:
+            coupling_map.append([q, lq])
+            coupling_map.append([lq, q])
+
+    for pair in itertools.permutations(free_qubits, 2):
+        coupling_map.append(pair)
+
     # convert qubits to architecture Nodes
+    characterisation = {}
     node_ers_dict = {Node(q_index): ers for q_index, ers in node_ers_dict.items()}
     link_ers_dict = {
         (Node(q_indices[0]), Node(q_indices[1])): ers
         for q_indices, ers in link_ers_dict.items()
     }
-    return Device(node_ers_dict, link_ers_dict, arc)
+
+    characterisation["NodeErrors"] = node_ers_dict
+    characterisation["EdgeErrors"] = link_ers_dict
+    characterisation["ReadoutErrors"] = readout_errors_dict
+    characterisation["GenericOneQubitQErrors"] = generic_single_qerrors_dict
+    characterisation["GenericTwoQubitQErrors"] = generic_2q_qerrors_dict
+    characterisation["Architecture"] = Architecture(coupling_map)
+
+    return characterisation
 
 
 def _sparse_to_qiskit_pauli(pauli: QubitPauliString, n_qubits: int) -> qk_Pauli:
     empty = np.zeros(n_qubits)
     q_pauli = qk_Pauli(empty, empty)
-    for q, p in pauli.map.items():
+    for q, p in pauli.to_dict().items():
         i = _default_q_index(q)
         if p in (Pauli.X, Pauli.Y):
             q_pauli._x[i] = True
